/**
 * Filename: Car.java (in org.openbandy.example)
 * This is part of the OpenBandy project.
 * 
 * www.openbandy.org
 */

package org.openbandy.example;

import java.util.Vector;

import org.openbandy.io.Serializable;
import org.openbandy.io.SerializableVector;
import org.openbandy.io.SerializationBuffer;
import org.openbandy.io.SerializationIdentifiers;
import org.openbandy.io.Serializer;

/**
 * An example business object that shows how to implement
 * the Serializable interface. This class contains attributes
 * as well as a vector of serializable objects.
 *
 * <br><br>
 * (c) Copyright P. Bolliger 2007, ALL RIGHTS RESERVED.
 *
 * @author Philipp Bolliger (philipp@bolliger.name)
 * @version 1.0
 */
public class Car implements Serializable{

	/* Storage serializer ID */
	protected int carId = SerializationIdentifiers.NO_ID;

	/* Attributes that will be serialized */
	protected String brand = "";
	protected int length = 0;

	/* Reference that will be serialized */
	protected Engine engine;

	/* A bandy serializable vector object to store
	 * the references for the tires we need */
	protected SerializableVector tires;


	/* ************ Constructor ************ */

	public Car(){
		/* ensure that no referenced object is null */
		engine = new Engine();
		tires = new SerializableVector();
		
		/* name the serializable vector 'tires',
		 * this way you could potentially have
		 * an arbitrary number of additional
		 * vectors in Car.java */
		tires.setSerializationName("tires");
	}

	
	/* ************ Getter and Setter Methods ************ */

	public String getBrand() {
		return brand;
	}

	public void setBrand(String brand) {
		this.brand = brand;
	}

	public int getLength() {
		return length;
	}

	public void setLength(int length) {
		this.length = length;
	}

	public Engine getEngine() {
		return engine;
	}

	public void setEngine(Engine engine) {
		this.engine = engine;
	}

	public SerializableVector getTires() {
		return tires;
	}

	public void setTires(SerializableVector tires) {
		this.tires = tires;
	}
	
	public void addTire(Tire tire){
		tires.add(tire);
	}

	
	/* ************ Methods for interface Serializable ************ */

	public void serialize(Serializer serializer, SerializationBuffer serializationBuffer) {
		serializationBuffer.container().addString("brand", brand);
		serializationBuffer.container().addInt("length", length);

		serializer.serializeObject(engine, serializationBuffer);
		serializer.serializeObject(tires, serializationBuffer);
	}

	public void deserialize(Serializer serializer, SerializationBuffer serializationBuffer) {
		brand = serializationBuffer.container().getString("brand");
		length = serializationBuffer.container().getInt("length");

		serializer.deserializeObject(engine, serializationBuffer);
		serializer.deserializeObject(tires, serializationBuffer);
	}

	public String getIdName() {
		return "carId";
	}

	public int getIdValue() {
		return carId;
	}

	public void setIdValue(int idValue) {
		carId = idValue;
	}

	public void getAttributeNames(Vector attributes) {
		attributes.addElement("brand");
		attributes.addElement("length");
	}

	public void getReferences(Vector references) {
		references.addElement(engine);
		references.addElement(tires);
	}

	public void getReferencedIdNames(Vector referenceNames) {
		referenceNames.addElement(engine.getIdName());
		referenceNames.addElement(tires.getSerializationName());
	}

	public String getLegibleString() {
		return "Car : "+brand+" "+length+" "+engine.getLegibleString()+" "+tires.getLegibleString();
	}

	public boolean isEqualShallow(Serializable serializableObject) throws ClassCastException {
		Car comparingCar = (Car)serializableObject;
		return ( (this.brand.equals(comparingCar.brand)) &&
				 (this.length == comparingCar.length) );
	}

	public boolean isEqual(Serializable serializableObject) throws ClassCastException {
		Car comparingCar = (Car)serializableObject;
		if(isEqualShallow(comparingCar)){
			return (this.tires.isEqual(comparingCar.tires) && this.engine.isEqual(comparingCar.engine) );
		}
		return false;
	}

	public Serializable getCopy() {
		Car clone = new Car();
		clone.carId = this.carId;
		clone.brand = this.brand;
		clone.length = this.length;
		clone.engine = (Engine)this.engine.getCopy();
		clone.tires = (SerializableVector)this.tires.getCopy();
		return clone;
	}

	public Serializable getNewInstance() {
		return new Car();
	}
}
