/**
 * Filename: StreamSerializerImpl.java (in org.openbandy.io.stream)
 * This file is part of the OpenBandy project.
 * 
 * www.openbandy.org
 * 
 */

package org.openbandy.io.stream;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

import org.kxml.Xml;
import org.kxml.parser.AbstractXmlParser;
import org.kxml.parser.ParseEvent;
import org.openbandy.io.Message;
import org.openbandy.io.Serializable;
import org.openbandy.io.SerializableFactory;
import org.openbandy.io.SerializableVector;
import org.openbandy.io.SerializationBuffer;
import org.openbandy.service.LogService;

import com.jcraft.jzlib.JZlib;
import com.jcraft.jzlib.ZInputStream;
import com.jcraft.jzlib.ZOutputStream;

/**
 * TODO describe purpose and usage
 * TODO check and optimze compression (vector!)
 * TODO check and fix ser/deser COMPRESSED_BEST (ZIP)
 *
 * <br><br>
 * (c) Copyright P. Bolliger 2007, ALL RIGHTS RESERVED.
 *
 * @author Philipp Bolliger (philipp@bolliger.name)
 * @version 1.0
 */
public class StreamSerializerImpl implements StreamSerializer{

	/* ****************** Methods for StreamSerializer ******************** */


	/* (non-Javadoc)
	 * @see org.openbandy.io.stream.StreamSerializer#getXmlStream(org.openbandy.io.Serializable, int)
	 */
	public ByteArrayOutputStream getXmlStream(Serializable serializable, int compression){

		/* create and setup new proxy object */
		StreamSerializationBuffer streamSerializationBuffer = new StreamSerializationBuffer();
		streamSerializationBuffer.setupForSerialization(serializable, compression);

		/* serialize object */
		serializeObject(serializable, streamSerializationBuffer);

		/* close writer */
		streamSerializationBuffer.closeWriter();

		/*  COMPRESSED_ZIP or COMPRESSED_BEST  */
		if((compression & StreamSerializer.COMPRESSED_ZIP) == StreamSerializer.COMPRESSED_ZIP){
			ByteArrayOutputStream zbaos = new ByteArrayOutputStream();
			try{
				ZOutputStream zOutputStream = new ZOutputStream(zbaos, JZlib.Z_BEST_COMPRESSION);
				zOutputStream.write(streamSerializationBuffer.outputStream.toByteArray());
				zOutputStream.close();
				return zbaos;
			}catch(OutOfMemoryError oome){
				LogService.error(this, "Could not compress : out of memory", oome);
			}catch(IOException ioe){
				LogService.error(this, "Could not compress", ioe);
			}
		}

		return streamSerializationBuffer.outputStream;
	}

	/* (non-Javadoc)
	 * @see org.openbandy.io.stream.StreamSerializer#getObjectFromXmlStream(java.io.ByteArrayInputStream)
	 */
	public Serializable getObjectFromXmlStream(ByteArrayInputStream inputStream){

		/* the object that will be serialized */
		Serializable serializable = null;

		/* create and setup new proxy object */
		StreamSerializationBuffer streamSerializationBuffer = new StreamSerializationBuffer();
		streamSerializationBuffer.setupForDeserialization(inputStream);


		/* ******************* NO_COMPRESSION ****************** */
		if(streamSerializationBuffer.streamType == StreamSerializer.COMPRESSION_NONE){

			/* create new object */
			String className = StreamSerializerHelper.readClassNameFromUncompressedStream(inputStream);
			serializable = SerializableFactory.getNewSerializableObject(className);

			if(serializable != null){

				/* prepare message object */
				if(serializable instanceof Message){
					((Message)serializable).prepare(this, streamSerializationBuffer);
				}

				/* setup parser */
				streamSerializationBuffer.setupXmlParser();

				/* deserialize object */
				deserializeObject(serializable, streamSerializationBuffer);
				return serializable;
			}
		}

		/* ******************* COMPRESSED_TOKEN ****************** */
		else if(streamSerializationBuffer.streamType == StreamSerializer.COMPRESSED_TOKEN){

			/* create new object */
			String className = StreamSerializerHelper.readClassNameFromTokenizedStream(inputStream);
			serializable = SerializableFactory.getNewSerializableObject(className);

			if(serializable != null){

				/* prepare message object */
				if(serializable instanceof Message){
					((Message)serializable).prepare(this, streamSerializationBuffer);
				}

				/* setup parser */
				streamSerializationBuffer.setupWbxmlParser(serializable);

				/* deserialize object */
				deserializeObject(serializable, streamSerializationBuffer);
				return serializable;
			}
		}

		/* ******************* COMPRESSED_ZIP or COMPRESSED_BEST ****************** */
		else if( (streamSerializationBuffer.streamType == StreamSerializer.COMPRESSED_ZIP)
				|| (streamSerializationBuffer.streamType == StreamSerializer.COMPRESSED_BEST)){

			ZInputStream zInputStream = new ZInputStream(inputStream);
			try{
				/* unzip stream */
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				int c = zInputStream.read();
				while(c != -1){
					baos.write(c);
					c = zInputStream.read();
				}

				/* deserialize from unzipped stream */
				return getObjectFromXmlStream(new ByteArrayInputStream(baos.toByteArray()));
			}catch(IOException ioe){
				ioe.printStackTrace();
			}
		}
		return null;
	}


	/* ****************** Methods for Serializer ******************** */

	public void serializeObject(Serializable serializable, SerializationBuffer serializationBuffer) {
		if(serializationBuffer != null){
			try{
				StreamSerializationBuffer streamSerializationBuffer = (StreamSerializationBuffer)serializationBuffer;

				if(!streamSerializationBuffer.isSetupUpForSerialzation()){
					LogService.warn(this, "Not setup for serialization. Check writer and container");
				}else{
					/* write start tag including the class name */
					streamSerializationBuffer.writer.startTag(StreamSerializerHelper.CLASSNAME_FLAG+serializable.getClass().getName());

					/* add id */
					String idName = serializable.getIdName();
					if(idName != null){
						streamSerializationBuffer.container().addInt(idName, serializable.getIdValue());
					}

					/* serialize object */
					serializable.serialize(this, streamSerializationBuffer);

					/* write end tag */
					streamSerializationBuffer.writer.endTag();
				}
			}catch(IOException ioe){
				LogService.error(this, ioe.getMessage(), ioe);
			}
			catch(ClassCastException cce){
				LogService.warn(this, "Cannot serialize: Wrong type of SerializationBuffer");
			}
		}
	}

	public void deserializeObject(Serializable serializable, SerializationBuffer serializationBuffer) {
		if(serializationBuffer != null){
			try{
				StreamSerializationBuffer streamSerializationBuffer = (StreamSerializationBuffer)serializationBuffer;

				if((serializable != null) && (streamSerializationBuffer.isSetupForDeserialization())){
					/* parse event */
					ParseEvent event = null;
					if(streamSerializationBuffer.lastReadEvent == null){
						event = readNextStartTagEvent(streamSerializationBuffer.parser);
					}else{
						event = streamSerializationBuffer.lastReadEvent;
						streamSerializationBuffer.lastReadEvent = null;
					}

					if(event != null){

						/* create container */
						StreamValueContainer container = new StreamValueContainer(event);
						streamSerializationBuffer.addContainer(container);

						/* set id */
						String idName = serializable.getIdName();
						if(idName != null){
							serializable.setIdValue(container.getInt(idName));
						}

						/* handle vector */
						if(serializable instanceof SerializableVector){
							if(container.getInt("size") != 0){
								/* set content class name */
								streamSerializationBuffer.lastReadEvent = readNextStartTagEvent(streamSerializationBuffer.parser);
								if(streamSerializationBuffer.lastReadEvent != null){
									((SerializableVector)serializable).setContainedObjectsClassName(streamSerializationBuffer.lastReadEvent.getName().substring(StreamSerializerHelper.CLASSNAME_FLAG.length()));
								}
							}
						}

						/* deserialize object */
						serializable.deserialize(this, streamSerializationBuffer);
					}
				}
			}catch(ClassCastException cce){
				LogService.warn(this, "Cannot serialize: Wrong type of SerializationBuffer");
			}
		}
	}

	public Serializable createContentObject(StreamSerializationBuffer streamSerializationBuffer) {
		String contentClassName = streamSerializationBuffer.getContentClassName();
		if( (contentClassName != null) && (contentClassName != "") ){
			return SerializableFactory.getNewSerializableObject(contentClassName);
		}
		return null;
	}

	/* ****************** Helper Methods ******************** */

	private ParseEvent readNextStartTagEvent(AbstractXmlParser parser){
		ParseEvent event = null;
		try{
			event = parser.read();
			while(event.getType() != Xml.START_TAG){
				event = parser.read();
				if(event.getType() == Xml.END_DOCUMENT){
					return null;
				}
			}
		}catch(IOException ioe){
			LogService.error(this, ioe.getMessage(), ioe);
		}
		return event;
	}
}